# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Graph manipulation utilities.

(dot generation adapted from pypy/translator/tool/make_dot.py)
"""

__docformat__ = "restructuredtext en"

__metaclass__ = type

import os.path as osp
import os
import sys
import tempfile
import codecs
import errno
from typing import Dict, List, Tuple, Optional, Set, TypeVar, Iterable


def escape(value):
    """Make <value> usable in a dot file."""
    lines = [line.replace('"', '\\"') for line in value.split("\n")]
    data = "\\l".join(lines)
    return "\\n" + data


def target_info_from_filename(filename):
    """Transforms /some/path/foo.png into ('/some/path', 'foo.png', 'png')."""
    basename = osp.basename(filename)
    storedir = osp.dirname(osp.abspath(filename))
    target = filename.split(".")[-1]
    return storedir, basename, target


class DotBackend:
    """Dot File backend."""

    def __init__(
        self,
        graphname,
        rankdir=None,
        size=None,
        ratio=None,
        charset="utf-8",
        renderer="dot",
        additionnal_param={},
    ):
        self.graphname = graphname
        self.renderer = renderer
        self.lines = []
        self._source = None
        self.emit("digraph %s {" % normalize_node_id(graphname))
        if rankdir:
            self.emit("rankdir=%s" % rankdir)
        if ratio:
            self.emit("ratio=%s" % ratio)
        if size:
            self.emit('size="%s"' % size)
        if charset:
            assert charset.lower() in ("utf-8", "iso-8859-1", "latin1"), (
                "unsupported charset %s" % charset
            )
            self.emit('charset="%s"' % charset)
        for param in sorted(additionnal_param.items()):
            self.emit("=".join(param))

    def get_source(self):
        """returns self._source"""
        if self._source is None:
            self.emit("}\n")
            self._source = "\n".join(self.lines)
            del self.lines
        return self._source

    source = property(get_source)

    def generate(self, outputfile=None, dotfile=None, mapfile=None):
        """Generates a graph file.

        :param outputfile: filename and path [defaults to graphname.png]
        :param dotfile: filename and path [defaults to graphname.dot]

        :rtype: str
        :return: a path to the generated file
        """
        import subprocess  # introduced in py 2.4

        name = self.graphname
        if not dotfile:
            # if 'outputfile' is a dot file use it as 'dotfile'
            if outputfile and outputfile.endswith(".dot"):
                dotfile = outputfile
            else:
                dotfile = "%s.dot" % name
        if outputfile is not None:
            storedir, basename, target = target_info_from_filename(outputfile)
            if target != "dot":
                pdot, dot_sourcepath = tempfile.mkstemp(".dot", name)
                os.close(pdot)
            else:
                dot_sourcepath = osp.join(storedir, dotfile)
        else:
            target = "png"
            pdot, dot_sourcepath = tempfile.mkstemp(".dot", name)
            ppng, outputfile = tempfile.mkstemp(".png", name)
            os.close(pdot)
            os.close(ppng)
        pdot = codecs.open(dot_sourcepath, "w", encoding="utf8")
        pdot.write(self.source)
        pdot.close()
        if target != "dot":
            if sys.platform == "win32":
                use_shell = True
            else:
                use_shell = False
            try:
                if mapfile:
                    subprocess.call(
                        [
                            self.renderer,
                            "-Tcmapx",
                            "-o",
                            mapfile,
                            "-T",
                            target,
                            dot_sourcepath,
                            "-o",
                            outputfile,
                        ],
                        shell=use_shell,
                    )
                else:
                    subprocess.call(
                        [self.renderer, "-T", target, dot_sourcepath, "-o", outputfile],
                        shell=use_shell,
                    )
            except OSError as e:
                if e.errno == errno.ENOENT:
                    e.strerror = "File not found: {0}".format(self.renderer)
                    raise
            os.unlink(dot_sourcepath)
        return outputfile

    def emit(self, line):
        """Adds <line> to final output."""
        self.lines.append(line)

    def emit_edge(self, name1, name2, **props):
        """emit an edge from <name1> to <name2>.
        edge properties: see http://www.graphviz.org/doc/info/attrs.html
        """
        attrs = ['%s="%s"' % (prop, value) for prop, value in props.items()]
        n_from, n_to = normalize_node_id(name1), normalize_node_id(name2)
        self.emit("%s -> %s [%s];" % (n_from, n_to, ", ".join(sorted(attrs))))

    def emit_node(self, name, **props):
        """emit a node with given properties.
        node properties: see http://www.graphviz.org/doc/info/attrs.html
        """
        attrs = ['%s="%s"' % (prop, value) for prop, value in props.items()]
        self.emit("%s [%s];" % (normalize_node_id(name), ", ".join(sorted(attrs))))


def normalize_node_id(nid):
    """Returns a suitable DOT node id for `nid`."""
    return '"%s"' % nid


class GraphGenerator:
    def __init__(self, backend):
        # the backend is responsible to output the graph in a particular format
        self.backend = backend

    # XXX doesn't like space in outpufile / mapfile
    def generate(self, visitor, propshdlr, outputfile=None, mapfile=None):
        # the visitor
        # the property handler is used to get node and edge properties
        # according to the graph and to the backend
        self.propshdlr = propshdlr
        for nodeid, node in visitor.nodes():
            props = propshdlr.node_properties(node)
            self.backend.emit_node(nodeid, **props)
        for subjnode, objnode, edge in visitor.edges():
            props = propshdlr.edge_properties(edge, subjnode, objnode)
            self.backend.emit_edge(subjnode, objnode, **props)
        return self.backend.generate(outputfile=outputfile, mapfile=mapfile)


class UnorderableGraph(Exception):
    pass


V = TypeVar("V")
_Graph = Dict[V, List[V]]


def ordered_nodes(graph: _Graph) -> Tuple[V, ...]:
    """takes a dependency graph dict as arguments and return an ordered tuple of
    nodes starting with nodes without dependencies and up to the outermost node.

    If there is some cycle in the graph, :exc:`UnorderableGraph` will be raised.

    Also the given graph dict will be emptied.
    """
    # check graph consistency
    cycles: List[List[V]] = get_cycles(graph)

    if cycles:
        bad_cycles = "\n".join([" -> ".join(map(str, cycle)) for cycle in cycles])
        raise UnorderableGraph("cycles in graph: %s" % bad_cycles)

    vertices = set(graph)
    to_vertices = set()

    for edges in graph.values():
        to_vertices |= set(edges)

    missing_vertices = to_vertices - vertices
    if missing_vertices:
        raise UnorderableGraph("missing vertices: %s" % ", ".join(missing_vertices))

    # order vertices
    order = []
    order_set = set()
    old_len = None

    while graph:
        if old_len == len(graph):
            raise UnorderableGraph("unknown problem with %s" % graph)

        old_len = len(graph)
        deps_ok = []

        for node, node_deps in graph.items():
            for dep in node_deps:
                if dep not in order_set:
                    break
            else:
                deps_ok.append(node)

        order.append(deps_ok)
        order_set |= set(deps_ok)

        for node in deps_ok:
            del graph[node]

    result = []

    for grp in reversed(order):
        result.extend(sorted(grp))

    return tuple(result)


def get_cycles(graph_dict: _Graph, vertices: Optional[Iterable] = None) -> List[List]:
    """given a dictionary representing an ordered graph (i.e. key are vertices
    and values is a list of destination vertices representing edges), return a
    list of detected cycles
    """
    if not graph_dict:
        return []

    result: List[List] = []
    if vertices is None:
        vertices = graph_dict.keys()

    for vertice in vertices:
        _get_cycles(graph_dict, [], set(), result, vertice)

    return result


def _get_cycles(
    graph_dict: _Graph, path: List, visited: Set, result: List[List], vertice: V
) -> None:
    """recursive function doing the real work for get_cycles"""
    if vertice in path:
        cycle = [vertice]

        for node in path[::-1]:
            if node == vertice:
                break

            cycle.insert(0, node)

        # make a canonical representation
        # mypy: error: Value of type variable "_LT" of "min" cannot be "V"
        # XXX I have no idea what this exactly mean here... we probably need to
        # XXX tell that "V" supports "lower than" type of comparison but how?
        start_from = min(cycle)  # type: ignore
        index = cycle.index(start_from)
        cycle = cycle[index:] + cycle[0:index]

        # append it to result if not already in
        if cycle not in result:
            result.append(cycle)
        return

    path.append(vertice)

    try:
        for node in graph_dict[vertice]:
            # don't check already visited nodes again
            if node not in visited:
                _get_cycles(graph_dict, path, visited, result, node)
                visited.add(node)

    except KeyError:
        pass

    path.pop()


def has_path(
    graph_dict: Dict[str, List[str]], fromnode: str, tonode: str, path: Optional[List[str]] = None
) -> Optional[List[str]]:
    """generic function taking a simple graph definition as a dictionary, with
    node has key associated to a list of nodes directly reachable from it.

    Return None if no path exists to go from `fromnode` to `tonode`, else the
    first path found (as a list including the destination node at last)
    """
    if path is None:
        path = []
    elif fromnode in path:
        return None
    path.append(fromnode)
    for destnode in graph_dict[fromnode]:
        if destnode == tonode or has_path(graph_dict, destnode, tonode, path):
            return path[1:] + [tonode]
    path.pop()
    return None



# -*- coding: utf-8 -*-
# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Extends the logging module from the standard library."""

__docformat__ = "restructuredtext en"

import os
import sys
import logging

from logilab.common.textutils import colorize_ansi


def set_log_methods(cls, logger):
    """bind standard logger's methods as methods on the class"""
    cls.__logger = logger
    for attr in ("debug", "info", "warning", "error", "critical", "exception"):
        setattr(cls, attr, getattr(logger, attr))


def xxx_cyan(record):
    if "XXX" in record.message:
        return "cyan"


class ColorFormatter(logging.Formatter):
    """
    A color Formatter for the logging standard module.

    By default, colorize CRITICAL and ERROR in red, WARNING in orange, INFO in
    green and DEBUG in yellow.

    self.colors is customizable via the 'color' constructor argument (dictionary).

    self.colorfilters is a list of functions that get the LogRecord
    and return a color name or None.
    """

    def __init__(self, fmt=None, datefmt=None, colors=None):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.colorfilters = []
        self.colors = {
            "CRITICAL": "red",
            "ERROR": "red",
            "WARNING": "magenta",
            "INFO": "green",
            "DEBUG": "yellow",
        }
        if colors is not None:
            assert isinstance(colors, dict)
            self.colors.update(colors)

    def format(self, record):
        msg = logging.Formatter.format(self, record)
        if record.levelname in self.colors:
            color = self.colors[record.levelname]
            return colorize_ansi(msg, color)
        else:
            for cf in self.colorfilters:
                color = cf(record)
                if color:
                    return colorize_ansi(msg, color)
        return msg


def set_color_formatter(logger=None, **kw):
    """
    Install a color formatter on the 'logger'. If not given, it will
    defaults to the default logger.

    Any additional keyword will be passed as-is to the ColorFormatter
    constructor.
    """
    if logger is None:
        logger = logging.getLogger()
        if not logger.handlers:
            logging.basicConfig()
    format_msg = logger.handlers[0].formatter._fmt
    fmt = ColorFormatter(format_msg, **kw)
    fmt.colorfilters.append(xxx_cyan)
    logger.handlers[0].setFormatter(fmt)


LOG_FORMAT = "%(asctime)s - (%(name)s) %(levelname)s: %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def get_handler(debug=False, syslog=False, logfile=None, rotation_parameters=None):
    """get an apropriate handler according to given parameters"""
    if os.environ.get("APYCOT_ROOT"):
        handler = logging.StreamHandler(sys.stdout)
    if debug:
        handler = logging.StreamHandler()
    elif logfile is None:
        if syslog:
            from logging import handlers

            handler = handlers.SysLogHandler()
        else:
            handler = logging.StreamHandler()
    else:
        try:
            if rotation_parameters is None:
                if os.name == "posix" and sys.version_info >= (2, 6):
                    from logging.handlers import WatchedFileHandler

                    handler = WatchedFileHandler(logfile)
                else:
                    handler = logging.FileHandler(logfile)
            else:
                from logging.handlers import TimedRotatingFileHandler

                handler = TimedRotatingFileHandler(logfile, **rotation_parameters)
        except IOError:
            handler = logging.StreamHandler()
    return handler


def get_threshold(debug=False, logthreshold=None):
    if logthreshold is None:
        if debug:
            logthreshold = logging.DEBUG
        else:
            logthreshold = logging.ERROR
    elif isinstance(logthreshold, str):
        logthreshold = getattr(logging, THRESHOLD_MAP.get(logthreshold, logthreshold))
    return logthreshold


def _colorable_terminal():
    isatty = hasattr(sys.__stdout__, "isatty") and sys.__stdout__.isatty()
    if not isatty:
        return False
    if os.name == "nt":
        try:
            from colorama import init as init_win32_colors
        except ImportError:
            return False
        init_win32_colors()
    return True


def get_formatter(logformat=LOG_FORMAT, logdateformat=LOG_DATE_FORMAT):
    if _colorable_terminal():
        fmt = ColorFormatter(logformat, logdateformat)

        def col_fact(record):
            if "XXX" in record.message:
                return "cyan"
            if "kick" in record.message:
                return "red"

        fmt.colorfilters.append(col_fact)
    else:
        fmt = logging.Formatter(logformat, logdateformat)
    return fmt


def init_log(
    debug=False,
    syslog=False,
    logthreshold=None,
    logfile=None,
    logformat=LOG_FORMAT,
    logdateformat=LOG_DATE_FORMAT,
    fmt=None,
    rotation_parameters=None,
    handler=None,
):
    """init the log service"""
    logger = logging.getLogger()
    if handler is None:
        handler = get_handler(debug, syslog, logfile, rotation_parameters)
    # only addHandler and removeHandler method while I would like a setHandler
    # method, so do it this way :$
    logger.handlers = [handler]
    logthreshold = get_threshold(debug, logthreshold)
    logger.setLevel(logthreshold)
    if fmt is None:
        if debug:
            fmt = get_formatter(logformat=logformat, logdateformat=logdateformat)
        else:
            fmt = logging.Formatter(logformat, logdateformat)
    handler.setFormatter(fmt)
    return handler


# map logilab.common.logger thresholds to logging thresholds
THRESHOLD_MAP = {
    "LOG_DEBUG": "DEBUG",
    "LOG_INFO": "INFO",
    "LOG_NOTICE": "INFO",
    "LOG_WARN": "WARNING",
    "LOG_WARNING": "WARNING",
    "LOG_ERR": "ERROR",
    "LOG_ERROR": "ERROR",
    "LOG_CRIT": "CRITICAL",
}
